#include <iostream.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <errno.h>
#include <X11/Xproto.h>

#include "constants.H"
#include "util.H"
#include "ClientMultiplexer.H"
#include "ServerMultiplexer.H"


#ifdef __EMX__

struct  sockaddr_un {
    u_short    sun_family;    /* socket family: AF_UNIX */
    char    sun_path[108];    /* path name (not used) */
};
#endif


#ifdef _AIX
#include <sys/select.h>
#endif /* _AIX */

#if defined(hpux) && !defined(RLIM_INFINITY)
/* HP-UX hides this define */
#define	RLIM_INFINITY	0x7fffffff
#endif


static void Usage(const char* const* argv);
static void Cleanup();
static void HandleSignal(int);
static int AwaitClientConnection(int portNum);
static int ConnectToServer(char* serverHost, int portNum);



// Maximum number of open file descriptors for this process
static unsigned int maxNumFDs = 0;


// Info about sockets used by the client proxy (global so that
// the cleanup signal handler can access it)
static char udomSocketPathname[100];
static int useUnixDomainSocket = 1;
sockaddr_in serverAddr;


// dxpc runs in client mode or server mode
enum ProxyMode { PROXY_CLIENT, PROXY_SERVER } proxyMode = PROXY_SERVER;


int
main(int argc, char** argv)
{
  (void)SaveCopyrightFromOptimizer();

  udomSocketPathname[0] = 0;

  unsigned int proxyPort = DEFAULT_PROXY_PORT;
  unsigned int displayNum = DEFAULT_DISPLAY_NUM;
  unsigned int statisticsLevel = 0;
  int useTCPSocket = 1;
  char* serverHost = NULL;
  // used to use getopt here, but it caused too many
  // portability problems
  for (int argi = 1; argi < argc; argi++)
  {
    char* nextArg = argv[argi];
    if (*nextArg == '-')
    {
      switch (nextArg[1])
      {
      case 'u':
	useUnixDomainSocket = 0;
	break;
      case 't':
	useTCPSocket = 0;
	break;
      case 'p':
	{
	  const char* arg = GetArg(argi, argc, argv);
	  if (arg == NULL)
	    Usage(argv);
	  else
	    proxyPort = atoi(arg);	
	}
	break;
      case 'd':
	{
	  const char* arg = GetArg(argi, argc, argv);
	  if (arg == NULL)
	    Usage(argv);
	  else
	    displayNum = atoi(arg);
	}
	break;
      case 's':
	{
	  const char* arg = GetArg(argi, argc, argv);
	  if (arg == NULL)
	    Usage(argv);
	  else
	    statisticsLevel = atoi(arg);
	}
	break;
      case 'v':
	PrintVersionInfo();
	exit(0);
      default:
	Usage(argv);
      }
    }
    else
    {
      if (serverHost != NULL)
	Usage(argv);
      else
	serverHost = nextArg;
    }
  }

  int xServerAddrFamily;
  sockaddr* xServerAddr;
  unsigned int xServerAddrLength;

  if (serverHost != NULL)
  {
    cout << "dxpc proxy running in CLIENT mode" << endl;
    proxyMode = PROXY_CLIENT;
  }
  else
  {
    cout << "dxpc proxy running in SERVER mode" << endl;

    // Keep Cleanup() from deleting the real X server's pipe
    // when we exit...
    useUnixDomainSocket = 0;

    // $DISPLAY is the X server for which we'll act as a proxy
    char* display = getenv("DISPLAY");
    if ((display == NULL) || (*display == 0))
    {
      cerr << "$DISPLAY is not set" << endl;
      Cleanup();
    }
    char* separator = strchr(display, ':');
    if ((separator == NULL) || !isdigit(*(separator + 1)))
    {
      cerr << "invalid DISPLAY '" << display << "'" << endl;
      Cleanup();
    }
    *separator = 0;
    int displayNum = atoi(separator + 1);
    if ((separator == display) || !strcmp(display, "unix"))
    {
      // UNIX domain port
      xServerAddrFamily = AF_UNIX;
      sockaddr_un* xServerAddrUNIX = new sockaddr_un;
      xServerAddrUNIX->sun_family = AF_UNIX;
      sprintf(udomSocketPathname, "/tmp/.X11-unix/X%d", displayNum);
      struct stat statInfo;
      if (stat(udomSocketPathname, &statInfo) == -1)
      {
#if (defined(__hpux) || defined(hpux)) && !defined(PGID_USE_PID)
	sprintf(udomSocketPathname, "/usr/spool/sockets/X11/%d", displayNum);
	if (stat(udomSocketPathname, &statInfo) == -1)
	{
#endif
	  cerr << "cannot open UNIX domain connection to X server" << endl;
	  Cleanup();
#if (defined(__hpux) || defined(hpux)) && !defined(PGID_USE_PID)
	}
#endif
      }
      strcpy(xServerAddrUNIX->sun_path, udomSocketPathname);
      xServerAddr = (sockaddr*)xServerAddrUNIX;
      //      xServerAddrLength = strlen(udomSocketPathname) + 2;
      xServerAddrLength = sizeof(sockaddr_un);
    }
    else
    {
      // TCP port
      xServerAddrFamily = AF_INET;
      int ipAddr;
      hostent* hostAddr = gethostbyname(display);
      if (hostAddr == NULL)
      {
	// on some UNIXes, gethostbyname doesn't accept IP addresses,
	// so try inet_addr:
	ipAddr = (int)inet_addr(display);
	if (ipAddr == -1)
	{
	  cerr << "Unknown host '" << display << "'" << endl;
	  Cleanup();
	}
      }
      else
	ipAddr = *(int*)hostAddr->h_addr_list[0];
      sockaddr_in* xServerAddrTCP = new sockaddr_in;
      xServerAddrTCP->sin_family = AF_INET;
      xServerAddrTCP->sin_port = htons(X_TCP_PORT + displayNum);
      xServerAddrTCP->sin_addr.s_addr = ipAddr;
      xServerAddr = (sockaddr*)xServerAddrTCP;
      xServerAddrLength = sizeof(sockaddr_in);
    }
  }

  // Increase the max # of open file descriptors for this process

  maxNumFDs = 0;
#if defined(RLIMIT_NOFILE)
  rlimit limits;
  if (getrlimit(RLIMIT_NOFILE, &limits) == 0)
  {
    if (limits.rlim_max == RLIM_INFINITY)
      maxNumFDs = 0;
    else
      maxNumFDs = (unsigned int)limits.rlim_max;
  }
#endif /* RLIMIT_NOFILE */

#if defined(_SC_OPEN_MAX)
  if (maxNumFDs == 0)
    maxNumFDs = sysconf(_SC_OPEN_MAX);
#endif

#if defined(FD_SETSIZE)
  if (maxNumFDs > FD_SETSIZE)
    maxNumFDs = FD_SETSIZE;
#endif /* FD_SETSIZE */

#if defined(RLIMIT_NOFILE)
  if (limits.rlim_cur < maxNumFDs)
  {
    limits.rlim_cur = (int)maxNumFDs;
    setrlimit(RLIMIT_NOFILE, &limits);
  }
#endif /* RLIMIT_NOFILE */

  if (maxNumFDs == 0)
  {
    cerr << "cannot determine number of available file descriptors, exiting!"
	 << endl;
    return 1;
  }

  // Install some signal handlers for graceful shutdown
  signal(SIGHUP, HandleSignal);
  signal(SIGINT, HandleSignal);
  signal(SIGTERM, HandleSignal);

  signal(SIGPIPE, (void (*)(int))SIG_IGN);


  // If running as client proxy, open sockets that mimic an
  // X display to which X clients can connect (e.g., unix:8
  // and <hostname>:8)
  int tcpFD = -1;
  int unixFD = -1;
  if (proxyMode == PROXY_CLIENT)
  {
    if (useTCPSocket)
    {
      // Open TCP socket for display
      tcpFD = socket(AF_INET, SOCK_STREAM, PF_UNSPEC);
      if (tcpFD == -1)
      {
	cerr << "socket() failed for TCP socket, errno=" << errno << endl;
	Cleanup();
      }
      int flag = 1;
      if (setsockopt(tcpFD, SOL_SOCKET, SO_REUSEADDR, (char*)&flag,
		     sizeof(flag)) < 0)
      {
	cerr << "setsockopt(SO_REUSEADDR) failed for TCP socket, errno=" <<
	  errno << endl;
      }
      sockaddr_in tcpAddr;
      tcpAddr.sin_family = AF_INET;
      unsigned int xPortTCP = X_TCP_PORT + displayNum;
      tcpAddr.sin_port = htons(xPortTCP);
      tcpAddr.sin_addr.s_addr = htonl(INADDR_ANY);
      if (bind(tcpFD, (sockaddr*)&tcpAddr, sizeof(tcpAddr)) == -1)
      {
	cerr << "bind() failed for TCP port " << xPortTCP <<
	  ", errno=" << errno << endl;
	Cleanup();
      }
      if (listen(tcpFD, 5) == -1)
      {
	cerr << "listen() failed for TCP port " << xPortTCP <<
	  ", errno=" << errno << endl;
	Cleanup();
      }
    }
    if (useUnixDomainSocket)
    {
      // Open UNIX domain socket for display
      unixFD = socket(AF_UNIX, SOCK_STREAM, PF_UNSPEC);
      if (unixFD == -1)
      {
	cerr << "socket() failed for UNIX domain socket, errno=" <<
	  errno << endl;
	Cleanup();
      }
      sockaddr_un unixAddr;
      unixAddr.sun_family = AF_UNIX;
      struct stat dirStat;
      if ((stat("/tmp/.X11-unix", &dirStat) == -1) && (errno == ENOENT))
      {
	mkdir("/tmp/.X11-unix", 0777);
	chmod("/tmp/.X11-unix", 0777);
      }
      sprintf(udomSocketPathname, "/tmp/.X11-unix/X%d", displayNum);
      strcpy(unixAddr.sun_path, udomSocketPathname);
      if (bind(unixFD, (sockaddr*)&unixAddr, strlen(udomSocketPathname) +2) ==
	  -1)
      {
	cerr << "bind() failed for UNIX domain socket " <<
	  udomSocketPathname << ", errno=" << errno << endl;
	Cleanup();
      }
      if (listen(unixFD, 5) == -1)
      {
	cerr << "listen() failed for UNIX domain socket " <<
	  udomSocketPathname << ", errno=" << errno << endl;
	Cleanup();
      }
    }
  }


  // Set up low-bandwidth connection between the proxies
  int proxyFD;
  if (proxyMode == PROXY_SERVER)
    proxyFD = AwaitClientConnection(proxyPort);
  else
    proxyFD = ConnectToServer(serverHost, proxyPort);
  if (proxyFD < 0)
    return 1;


  //// Create multiplexer
  Multiplexer* multiplexer;
  if (proxyMode == PROXY_SERVER)
    multiplexer = new ServerMultiplexer(proxyFD, xServerAddrFamily,
					xServerAddr, xServerAddrLength,
					statisticsLevel);
  else
    multiplexer = new ClientMultiplexer(proxyFD, statisticsLevel);


  // Loop endlessly, reading from all of the open file descriptors
  for (;;)
  {
    fd_set readSet;
    FD_ZERO(&readSet);
    FD_SET(proxyFD, &readSet);
    unsigned int numFDsToSelect = proxyFD + 1;
    if (proxyMode == PROXY_CLIENT)
    {
      if (useTCPSocket)
      {
	FD_SET(tcpFD, &readSet);
	if (tcpFD >= numFDsToSelect)
	  numFDsToSelect = tcpFD + 1;
      }
      if (useUnixDomainSocket)
      {
	FD_SET(unixFD, &readSet);
	if (unixFD >= numFDsToSelect)
	  numFDsToSelect = unixFD + 1;
      }
    }
    multiplexer->setSelectFDs(&readSet, numFDsToSelect);

    timeval delay;
    delay.tv_sec = 600;
    delay.tv_usec = 0;
    // PGID_USE_PID, defined in <sys/types.h>, is specific to HP-UX 10.x
#if (defined(__hpux) || defined(hpux)) && !defined(PGID_USE_PID)
    int result = select(numFDsToSelect, (int*)&readSet, NULL, NULL, &delay);
#else
    int result = select(numFDsToSelect, &readSet, NULL, NULL, &delay);
#endif
    if (result == -1)
    {
      if (errno == EINTR)
	continue;
      cerr << "select() failed, errno=" << errno << endl;
      Cleanup();
    }
    for (unsigned int j = 0; j < numFDsToSelect; j++)
    {
      if (!(FD_ISSET(j, &readSet)))
	continue;
      if (proxyMode == PROXY_CLIENT)
      {
	if ((j == tcpFD) && useTCPSocket)
	{
	  sockaddr_in newAddr;
#if defined(DXPC_ACCEPT_IS_SIZE_T)
	  size_t addrLen = sizeof(sockaddr_in);
#else
	  int addrLen = sizeof(sockaddr_in);
#endif /* defined(DXPC_ACCEPT_IS_SIZE_T) */
	  int newFD = accept(tcpFD, (sockaddr*)&newAddr, &addrLen);
	  if (newFD == -1)
	  {
	    cerr << "accept() failed, errno=" << errno << endl;
	    Cleanup();
	  }
	  multiplexer->createNewConnection(newFD);
	  continue;
	}
	if ((j == unixFD) && useUnixDomainSocket)
	{
	  sockaddr_un newAddr;
#if defined(DXPC_ACCEPT_IS_SIZE_T)
	  size_t addrLen = sizeof(sockaddr_un);
#else
	  int addrLen = sizeof(sockaddr_un);
#endif /* defined(DXPC_ACCEPT_IS_SIZE_T) */
	  int newFD = accept(unixFD, (sockaddr*)&newAddr, &addrLen);
	  if (newFD == -1)
	  {
	    cerr << "accept() failed, errno=" << errno << endl;
	    Cleanup();
	  }
	  multiplexer->createNewConnection(newFD);
	  continue;
	}
      }

      if (!multiplexer->handleSelect(j))
	Cleanup();
    }
  }

  return 0;
}


static void
Usage(const char* const* argv)
{
  cerr << "usage:\n (client)\n  " << argv[0] <<
    " [-d display_num] [-p port_num] [-u] [-s (1|2)] [-v] server_hostname\n" <<
    " (server)\n  " << argv[0] << " [-p port_num] [-s (1|2)] [-v]\n" <<  endl;
  exit(1);
}


static void
Cleanup()
{
  cout << "Closing all file descriptors and shutting down..." << endl;
  if (useUnixDomainSocket)
    unlink(udomSocketPathname);
  for (int i = 0; i < maxNumFDs; i++)
    (void)close(i);
  exit(0);
}


static void
HandleSignal(int)
{
  Cleanup();
}


// Open TCP socket to listen for client proxy; block until client
// proxy connects, then close listener socket and return FD of socket
// on which client proxy is connected
//
static int
AwaitClientConnection(int portNum)
{
  int proxyFD = socket(AF_INET, SOCK_STREAM, 0);
  if (proxyFD == -1)
  {
    cerr << "socket() failed for TCP socket, errno=" << errno << endl;
    Cleanup();
  }
  int flag = 1;
  if (setsockopt(proxyFD, SOL_SOCKET, SO_REUSEADDR, (char*)&flag,
		 sizeof(flag)) < 0)
  {
    cerr << "setsockopt(SO_REUSEADDR) failed for proxy port, errno=" <<
      errno << endl;
  }

  sockaddr_in tcpAddr;
  tcpAddr.sin_family = AF_INET;
  tcpAddr.sin_port = htons(portNum);
  tcpAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(proxyFD, (sockaddr*)&tcpAddr, sizeof(tcpAddr)) == -1)
  {
    cerr << "bind() failed for TCP port " << portNum <<
      ", errno=" << errno << endl;
    Cleanup();
  }
  if (listen(proxyFD, 1) == -1)
  {
    cerr << "listen() failed for TCP port " << portNum <<
      ", errno=" << errno << endl;
    Cleanup();
  }

  for (;;)
  {
    fd_set readSet;
    FD_ZERO(&readSet);
    FD_SET(proxyFD, &readSet);
#if (defined(__hpux) || defined(hpux)) && !defined(PGID_USE_PID)
    int result = select(proxyFD + 1, (int*)&readSet, NULL, NULL, NULL);
#else
    int result = select(proxyFD + 1, &readSet, NULL, NULL, NULL);
#endif
    if (result == -1)
    {
      if (errno == EINTR)
	continue;
      cerr << "select() failed, errno=" << errno << endl;
      Cleanup();
    }
    if (FD_ISSET(proxyFD, &readSet))
    {
      sockaddr_in newAddr;
#if defined(DXPC_ACCEPT_IS_SIZE_T)
      size_t addrLen = sizeof(sockaddr_in);
#else
      int addrLen = sizeof(sockaddr_in);
#endif /* defined(DXPC_ACCEPT_IS_SIZE_T) */
      int newFD = accept(proxyFD, (sockaddr*)&newAddr, &addrLen);
      if (newFD == -1)
      {
	cerr << "accept() failed, errno=" << errno << endl;
	Cleanup();
      }
      int flag = 1;
      setsockopt(newFD, IPPROTO_TCP, TCP_NODELAY, (char*)&flag, sizeof(int));
      cout << "client proxy connected\nready" << endl;
      close(proxyFD);
      return newFD;
    }
  }
}


// Connect to server proxy.  If successful, return FD of connection;
// if unsuccessful, return -1
//
static int
ConnectToServer(char* serverHost, int portNum)
{
  int serverIPAddr;
  hostent* hostAddr = gethostbyname(serverHost);
  if (hostAddr == NULL)
  {
    // on some UNIXes, gethostbyname doesn't accept IP addresses,
    // so try inet_addr:
    serverIPAddr = (int)inet_addr(serverHost);
    if (serverIPAddr == -1)
    {
      cerr << "Unknown host '" << serverHost << "'" << endl;
      exit(1);
    }
  }
  else
    serverIPAddr = *(int*)hostAddr->h_addr_list[0];

  int serverProxyFD = socket(AF_INET, SOCK_STREAM, PF_UNSPEC);
  if (serverProxyFD == -1)
  {
    cerr << "socket() failed, errno=" << errno << endl;
    exit(1);
  }
  int flag = 1;
  if (setsockopt(serverProxyFD, SOL_SOCKET, SO_REUSEADDR, (char*)&flag,
		 sizeof(flag)) < 0)
  {
    cerr << "setsockopt(SO_REUSEADDR) failed for proxy port, errno=" <<
      errno << endl;
  }

  cout << "trying to connect to server proxy..." << endl;
  sockaddr_in addr;
  addr.sin_family = AF_INET;
  addr.sin_port = htons(portNum);
  addr.sin_addr.s_addr = serverIPAddr;
  if (connect(serverProxyFD, (sockaddr*)&addr, sizeof(sockaddr_in)) == -1)
  {
    cerr << "connect() failed, errno=" << errno << endl;
    close(serverProxyFD);
    exit(1);
  }
  flag = 1;
  // setsockopt(serverProxyFD, IPPROTO_TCP, TCP_NODELAY,
  // 	     (char*)&flag, sizeof(int));

  cout << "connected to server proxy\nready" << endl;
  return serverProxyFD;
}
